"""
This module contains exceptions native to django-rest-easy.
"""

__all__ = ['RestEasyException']


class RestEasyException(Exception):
    """
    Default django-rest-easy exception class.
    """
